<footer class="md-footer">
    <div class="md-footer">
        {% if page.previous_page or page.next_page %}
        <nav class="md-footer__inner md-grid" aria-label="{{ lang.t('footer.title') }}">
            {% if page.previous_page %}
            <a href="{{ page.previous_page.url | url }}" class="md-footer__link md-footer__link--prev"
                rel="prev">
                <div class="md-footer__button md-icon">
                    {% include ".icons/material/arrow-left.svg" %}
                </div>
                <div class="md-footer__title">
                    <div class="md-ellipsis">
                        <span class="md-footer__direction">
                            {{ lang.t("footer.previous") }}
                        </span>
                        {{ page.previous_page.title }}
                    </div>
                </div>
            </a>
            {% endif %}
            {% if page.next_page %}
            <a href="{{ page.next_page.url | url }}" class="md-footer__link md-footer__link--next" rel="next">
                <div class="md-footer__title">
                    <div class="md-ellipsis">
                        <span class="md-footer__direction">
                            {{ lang.t("footer.next") }}
                        </span>
                        {{ page.next_page.title }}
                    </div>
                </div>
                <div class="md-footer__button md-icon">
                    {% include ".icons/material/arrow-right.svg" %}
                </div>
            </a>
            {% endif %}
        </nav>
    </div>
    {% endif %}
    <div class="md-footer-meta md-typeset">
        <div class="md-footer-meta__inner md-grid">
            <div class="md-copyright">
                Made with
                <a href="https://squidfunk.github.io/mkdocs-material/" target="_blank" rel="noopener">
                    Material for MkDocs
                </a>
            </div>
        </div>
        <div class="md-footer-meta__inner md-grid">
            <div class="md-copyright">
                Copyright © 2021 Eclipse Foundation<br>
                <a href="https://www.eclipse.org/legal/privacy.php" target="_blank">Privacy policy</a><br>
                <a href="https://www.eclipse.org/legal/termsofuse.php" target="_blank">Terms of use</a><br>
                <a href="https://www.eclipse.org/legal/copyright.php" target="_blank">Copyright</a><br>
                <a href="https://www.eclipse.org/legal/" target="_blank">Legal</a><br>
                <a href="https://www.eclipse.org/security/" target="_blank">Report a safety or security
                    vulnerability</a>
            </div>
        </div>
        <div class="md-footer-meta__inner md-grid">
            <div class="md-copyright">
                Sponsored by<br>
                <a href="https://www.eclipse.org/" target="_blank" rel="noopener" title="Eclipse"><img
                        src="{{ base_url }}/images/eclipse-logo.svg" width="100px"></img></a><br>
                <a href="https://www.bosch.com/" target="_blank" rel="noopener" title="Robert Bosch GmbH"><img
                        src="{{ base_url }}/images/bosch-logo.png" width="100px"></img></a><br>
                <a href="https://www.apex.ai/" target="_blank" rel="noopener" title="Apex.AI Inc."><img
                        src="{{ base_url }}/images/apex-logo.webp" width="100px"></img></a><br>
            </div>
        </div>
        <div class="md-footer-meta__inner md-grid">
            <div class="md-social">
                <a href="https://github.com/eclipse-iceoryx/iceoryx" target="_blank" rel="noopener" title="GitHub"
                    class="md-social__link">
                    {% include ".icons/material/github.svg" %}
                </a>
                <a href="https://gitter.im/eclipse/iceoryx" target="_blank" rel="noopener" title="Gitter.im Chat"
                    class="md-social__link">
                    {% include ".icons/material/chat.svg" %}
                </a>
                <a href="https://github.com/eclipse-iceoryx/iceoryx/blob/v2.0.0/GOVERNANCE.md#bi-weekly-meetup"
                    target="_blank" rel="noopener" title="Developer Meetup" class="md-social__link">
                    {% include ".icons/material/calendar.svg" %}
                </a>
                <a href="https://accounts.eclipse.org/mailing-list/iceoryx-dev" target="_blank" rel="noopener"
                    title="Mailing List" class="md-social__link">
                    {% include ".icons/material/email-multiple.svg" %}
                </a>
            </div>
        </div>
    </div>
</footer>