// Copyright (c) 2020 by Robert Bosch GmbH. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// SPDX-License-Identifier: Apache-2.0

#ifndef IOX_BINDING_C_TEST_HPP
#define IOX_BINDING_C_TEST_HPP

#include <gmock/gmock.h>
#include <gtest/gtest.h>

#include <numeric>

namespace iox_test_binding_c
{
template <typename T>
T maxUnderlyingCEnumValue()
{
// ignore the warning since we would like to test the behavior of an invalid enum value
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
    return static_cast<T>(std::numeric_limits<std::underlying_type_t<T>>::max());
#pragma GCC diagnostic pop
};
}; // namespace iox_test_binding_c

#endif // IOX_BINDING_C_TEST_HPP
