// Copyright (c) 2019 by Robert Bosch GmbH. All rights reserved.
// Copyright (c) 2021 by Apex.AI inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// SPDX-License-Identifier: Apache-2.0

#include "test.hpp"

using namespace ::testing;
using ::testing::_;

// global argc and argv needed by the argv_inspection test
int g_argc;
char** g_argv;

int main(int argc, char* argv[])
{
    ::testing::InitGoogleTest(&argc, argv);

    g_argc = argc;
    g_argv = argv;

    return RUN_ALL_TESTS();
}
