// Copyright (c) 2020 by Robert Bosch GmbH. All rights reserved.
// Copyright (c) 2021 by Apex.AI Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//
// SPDX-License-Identifier: Apache-2.0

#include "iceoryx_hoofs/platform/getopt.hpp"
#include "iceoryx_hoofs/platform/windows.hpp"

#include <cstdio>

int optind;
int opterr;
int optout;

int getopt_long(int argc, char* const[], const char*, const struct option*, int*)
{
    if (argc > 1)
    {
        fprintf(stderr, "%s is not implemented in windows!\n", __PRETTY_FUNCTION__);
        fprintf(stderr, "command line arguments are not supported in windows\n");
    }
    return -1;
}
